/** 
 *   FIXME ....
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ???
 *
 **/

#include <errno.h> 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "socket_srv.h"
#include "my_stdio.h"

#define COMMAND_LINE_LEGEND 80

struct client {
	int fd;
	int transfer;
	int alive;
	char cmd[COMMAND_LINE_LEGEND];
	int cmd_ind;
	char *mesg;
	struct client *prev;
	struct client *next;
};


struct client *init_client()
{
	struct client *cl;

	cl = (struct client *) malloc(sizeof(struct client));
	cl->fd = -1;
	cl->transfer = 0;
	cl->alive = 1;
	cl->cmd_ind = 0;
	cl->mesg = NULL;
	cl->next = cl->prev = NULL;
	
	return cl;
}


int print_client_data(char *buff, struct client *cl, int all)
{
	struct client *clp = cl;
	int l;

	l = strlen(buff)-1;
	if (buff[l] == '\n') buff[l] = 0;
	while(clp->next != NULL){
		if ( clp->fd && 
		     (clp->transfer || all) && 
		     clp->alive) {
			if( !fd_printf(clp->fd,"%s\n", buff) ){
				perror("write client");
				clp->alive = 0;
			}
		}
		clp = clp->next;
	}
	
	return 0;
}


int command_from_client(struct client *clp)
{
	char c;
	char *cmd;
	
	if ( read(clp->fd, &c, 1) <= 0) { /* Ops... die */
		printf("read client fail\n");
		close(clp->fd);
		clp->alive = 0;
		return -1;
	}
	if (c != '\n'){
		*(clp->cmd+clp->cmd_ind) = c;
	}
	else{
		*(clp->cmd+clp->cmd_ind) = 0;
	}
	clp->cmd_ind++;
	if (clp->cmd_ind == COMMAND_LINE_LEGEND) clp->cmd_ind = 0;
	if ( c == '\n') {
		clp->cmd_ind = 0;
		cmd = stripwhite(clp->cmd);
		switch(cmd[0]){
		case 'S':
		case 's': clp->transfer = 1;
			break;
		case 'T':
		case 't': clp->transfer = 0;
			break;
		case 'Q':
		case 'q':
			close(clp->fd);
			clp->alive = 0;
			break;
		case 'M':
		case 'm':
			if (cmd[1] == ':') {
				clp->mesg = stripwhite(cmd+2);
				return 2;
			}
			break;
		case '?':
			fd_printf(clp->fd,
				  "# s: start\n# t: stop\n# q: quit\n# m:: message to all\n# ?: help\n");
			break;
		}
	}

	return 0;
}


#if 0
void intHandler (int signo) 
{
      fprintf(stderr, "exit on signal %d\n", signo);
      exit(0);
}
#endif



int track_transfer_loop_main( int fd_pipe, int fd_socket )
{
	struct client *cl, *clp, *cld;
	fd_set rfds;
	struct timeval tv;
	int retval;
	char buff[PIPE_BUFFER_LEN];
	int l;
	struct sockaddr isa;
	int addrlen;

	addrlen = sizeof(isa); /* find socket's address for accept() */
	getsockname(fd_socket, (struct sockaddr *)&isa, &addrlen);

#if 0
	if (signal(SIGSEGV, intHandler) == SIG_ERR) {
		perror("Signal Failed (SIGSEGV).");
		exit(0);
	}
#endif
	cl = init_client();
	for(;1;){
		clp = cl ;
		FD_ZERO(&rfds);
		FD_SET(fd_socket, &rfds);
		if (fd_pipe >= 0) FD_SET(fd_pipe, &rfds);
		while(clp->next != NULL) {
			if (clp->fd && clp->alive ){ 
				FD_SET(clp->fd, &rfds);
				clp = clp->next;
			}
			else if ( !clp->alive ) { 
                                /* OK: client dead, cleaning */ 
				cld = clp;
				if (clp->prev != NULL){
					clp = cld->prev;
					clp->next = cld->next;
					cld->next->prev = clp;
					clp = clp->next;
				}else {
					clp = cl = clp->next;
					cl->prev = NULL;
				}
				free(cld);		
			}
		}
		tv.tv_sec = 10;
		tv.tv_usec = 0;
		/* FIXME: 50 ? implement this dynamic */
		retval = select(50, &rfds, NULL, NULL, &tv);
		if (retval) {
			clp = cl ;
			while(clp->next != NULL) {
				if( FD_ISSET(clp->fd, &rfds) && clp->alive ) {
					if (command_from_client(clp) == 2 &&
					    clp->mesg != NULL ) {
						print_client_data(clp->mesg,
								  cl, 1);
						clp->mesg = NULL;
					}
				}
				clp = clp->next;
			}
			if (FD_ISSET(fd_socket, &rfds)) {
				clp = cl ;
				while(clp->next != NULL) clp = clp->next;
				clp->fd = accept(fd_socket, 
						 (struct sockaddr *)&isa, 
						 &addrlen); 
				if( clp->fd ){
					clp->alive = 1;
					printf("connected\n");
					clp->next = init_client();
					clp->next->prev = clp;
				}
			}
			if ((fd_pipe >= 0) && FD_ISSET(fd_pipe, &rfds)) {
				l = read(fd_pipe, buff, PIPE_BUFFER_LEN);
				if (l > 0) {
					buff[l-1] = 0;
					print_client_data(buff, cl, 0);
				}
			}
		}
	}

	return 0;
}




int track_transfer_loop_unix( int fd_pipe, char *sock_name )
{
	int fd_socket;

	if ( (fd_socket = socket_init_unix( sock_name )) < 0) {
		return -1;
	}

	return track_transfer_loop_main( fd_pipe, fd_socket );
}


int track_transfer_loop( int fd_pipe, int port )
{
	int fd_socket;

	if ( (fd_socket = socket_init( port )) < 0) {
		return -1;
	}

	return track_transfer_loop_main( fd_pipe, fd_socket);
}


